// автор: Роман Тезиков

import io.netty.bootstrap.ServerBootstrap
import io.netty.buffer.{ByteBuf, Unpooled}
import io.netty.channel.nio.NioEventLoopGroup
import io.netty.channel.socket.SocketChannel
import io.netty.channel.socket.nio.NioServerSocketChannel
import io.netty.channel.*
import io.netty.util.CharsetUtil

import scala.util.Try

class Server(val port: Int) {
  var concurrentClientCount = 0

  @throws[Exception]
  def run(): ChannelFuture = {
    val bossGroup = new NioEventLoopGroup()
    val workerGroup = new NioEventLoopGroup()
    try {
      val b = new ServerBootstrap()
      b.group(bossGroup, workerGroup)
        .channel(classOf[NioServerSocketChannel])
        .childHandler(new ChannelInitializer[SocketChannel]() {
          @throws[Exception]
          override def initChannel(ch: SocketChannel): Unit = {
            ch.pipeline().addLast(new ServerHandler())
          }
        })
        .option(
          ChannelOption.SO_BACKLOG.asInstanceOf[ChannelOption[Any]],
          128
        )
        .childOption(ChannelOption.SO_KEEPALIVE.asInstanceOf[ChannelOption[Any]],
          true
        )
      val f = b.bind(port).sync()

      f.channel().closeFuture().sync()
    } finally {
      workerGroup.shutdownGracefully()
      bossGroup.shutdownGracefully()
    }
  }

  class ServerHandler extends ChannelInboundHandlerAdapter {
    override def channelActive(ctx: ChannelHandlerContext): Unit = {
      concurrentClientCount += 1
      println("$concurrentClientCount concurrent clients are connected\n")
    }

    var dataForProcessing = ""

    override def channelRead(ctx: ChannelHandlerContext, msg: Any): Unit = {
      try {
        //мы склеиваем остаток прошлого запроса с тем, что пришло
        val query = dataForProcessing + msg.asInstanceOf[ByteBuf]
          .toString(CharsetUtil.UTF_8)

        //а здесь расклеиваем на отдельные запросы
        var msgs = query.split("\r\n").toList

        //мы проверяем, чем окончивается запрос, и если это не \r\n, значит запрос неполный, и его нужно запомнить
        //пока не придет его окончание
        if (!query.endsWith("\r\n")) {
          dataForProcessing = msgs.last
          println("data for processing: " + dataForProcessing)
          msgs = msgs.dropRight(1)
          println("msgs: " + msgs)
        } else {
          dataForProcessing = ""
        }

        for (m <- msgs) {
          val words = m.trim.split(" +").toList

          val result = words match {
            case head :: operation :: left :: right :: _
              if head == Constants.CalcString => {
              Try {
                Calculator(operation, left.toLong, right.toLong)
              }.map {
                calculator => Constants.Okay(calculator.result)
              } getOrElse Constants.Error
            }
            case _ => Constants.Error
          }
          print(words)
          print(": ")
          print(result)
          ctx.write(Unpooled.copiedBuffer(result, CharsetUtil.UTF_8))
        }
        ctx.flush()
      } catch {
        case _: Exception =>
          ctx.write(Unpooled.copiedBuffer("ERR\r\n", CharsetUtil.UTF_8))
          ctx.flush()
      } finally {
        msg.asInstanceOf[ByteBuf].release()
      }
    }

    override def channelInactive(ctx: ChannelHandlerContext): Unit = {
      concurrentClientCount -= 1
    }
  }

}

case class Calculator(operation: String, left: Long, right: Long) {
  val result: Long = operation match {
    case "+" => left + right
    case "-" => left - right
    case "*" => left * right
    case "/" => left / right
  }
}

object Main {
  @throws[Exception]
  def main(args: Array[String]): Unit = {
    new Server(28563).run()
  }
}

object Constants {
  val Error = "ERR \n"
  val CalcString = "CALC"

  def Okay(result: Long) = s"$result\r\n"
}


